package com.bergerkiller.bukkit.nolagg;

import java.util.HashMap;
import java.util.Map;

import org.bukkit.block.BlockFace;

public class ChunkCoordIndexer {
	private static Map<ChunkCoordIndexer, Integer> indices = new HashMap<ChunkCoordIndexer, Integer>();
	public static int get(final int dx, final int dz, final BlockFace direction) {
		ChunkCoordIndexer indexer = new ChunkCoordIndexer(dx, dz, direction);
		Integer value = indices.get(indexer);
		if (value != null) return value;
		indexer.offset();
		indices.put(indexer, indexer.index);
		return indexer.index;
	}
	
	private final int dx, dz;
	private final BlockFace direction;
	private int index = -1;
	private ChunkCoordIndexer(final int dx, final int dz, final BlockFace direction) {
		this.dx = dx;
		this.dz = dz;
		this.direction = direction;
	}
	
	public boolean equals(Object object) {
		if (object == this) return true;
		if (object instanceof ChunkCoordIndexer) {
			ChunkCoordIndexer other = (ChunkCoordIndexer) object;
			return other.dx == this.dx && other.dz == this.dz && other.direction == this.direction;
		}
		return false;
	}
	public int hashcode() {
		return this.dx * 20 + this.dz * 800 + this.direction.ordinal();
	}
	
	private static class MoveMod {
		private MoveMod(BlockFace direction, boolean right) {
			this.direction = direction;
			this.right = right;
		}
		
		public BlockFace direction;
		public boolean right;
		
		public void next(int dx, int dz, int limit) {
			if ((dx == limit || dx == -limit) && (dz == limit || dz == -limit)) {
				if (this.right) {
					if (direction == BlockFace.NORTH) {
						direction = BlockFace.EAST;
					} else if (direction == BlockFace.EAST) {
						direction = BlockFace.SOUTH;
					} else if (direction == BlockFace.SOUTH) {
						direction = BlockFace.WEST;
					} else if (direction == BlockFace.WEST) {
						direction = BlockFace.NORTH;
					}
				} else {
					if (direction == BlockFace.NORTH) {
						direction = BlockFace.WEST;
					} else if (direction == BlockFace.WEST) {
						direction = BlockFace.SOUTH;
					} else if (direction == BlockFace.SOUTH) {
						direction = BlockFace.EAST;
					} else if (direction == BlockFace.EAST) {
						direction = BlockFace.NORTH;
					}
				}
			}
		}
		
		public static MoveMod[] get(BlockFace direction) {
			MoveMod[] mods = new MoveMod[2];
			if (direction == BlockFace.NORTH) {
				mods[0] = new MoveMod(BlockFace.WEST, false);
				mods[1] = new MoveMod(BlockFace.EAST, true);
			} else if (direction == BlockFace.SOUTH) {
				mods[0] = new MoveMod(BlockFace.WEST, true);
				mods[1] = new MoveMod(BlockFace.EAST, false);
			} else if (direction == BlockFace.EAST) {
				mods[0] = new MoveMod(BlockFace.NORTH, false);
				mods[1] = new MoveMod(BlockFace.SOUTH, true);
			} else if (direction == BlockFace.WEST) {
				mods[0] = new MoveMod(BlockFace.NORTH, true);
				mods[1] = new MoveMod(BlockFace.SOUTH, false);
			} else if (direction == BlockFace.NORTH_EAST) {
				mods[0] = new MoveMod(BlockFace.WEST, false);
				mods[1] = new MoveMod(BlockFace.SOUTH, true);
			} else if (direction == BlockFace.SOUTH_EAST) {
				mods[0] = new MoveMod(BlockFace.WEST, true);
				mods[1] = new MoveMod(BlockFace.NORTH, false);
			} else if (direction == BlockFace.SOUTH_WEST) {
				mods[0] = new MoveMod(BlockFace.NORTH, true);
				mods[1] = new MoveMod(BlockFace.EAST, false);
			} else if (direction == BlockFace.NORTH_WEST) {
				mods[0] = new MoveMod(BlockFace.SOUTH, false);
				mods[1] = new MoveMod(BlockFace.EAST, true);
			}
			return mods;
		}	
	}
	
	private boolean offset(final int newdx, final int newdz) {
		this.index++;
		return newdx == this.dx && newdz == this.dz;
	}
	
	private boolean offsetLayer(final int layer, final double factor) {
		int count = (int) (layer * factor) + 1;
		//get modifiers from direction
		MoveMod[] mods = MoveMod.get(direction);
		//Get the chunk to start at
		int startx = direction.getModX() * layer;
		int startz = direction.getModZ() * layer;
		//Send starter chunk
		if (this.offset(startx, startz)) return true;
		//Peel
		int x1 = startx;
		int z1 = startz;
		int x2 = startx;
		int z2 = startz;
		while (true) {
			if (--count == 0) return false;
			//offset the chunks
			x1 += mods[0].direction.getModX();
			z1 += mods[0].direction.getModZ();
			x2 += mods[1].direction.getModX();
			z2 += mods[1].direction.getModZ();
			//mod update
			mods[0].next(x1, z1, layer);
			mods[1].next(x2, z2, layer);
			//got till the end?
			if (x1 == x2 && z1 == z2) {
				return this.offset(x1, z1);
			} else {
				if (this.offset(x1, z1)) return true;
				if (this.offset(x2, z2)) return true;
			}
		}
	}
	
	private void offset() {
		//main chunk
		if (this.offset(0, 0)) return;
		
		final int threshold1 = 2; //to this layer full layers are sent, after half
		final int threshold2 = 4; //at this layer less than half are sent
		
		//full layers
		for (int layer = 1; layer < threshold1; layer++) {
			if (this.offsetLayer(layer, 4)) return;
		}
		
		//half layers
		for (int layer = threshold1; layer <= threshold2; layer++) {
			if (this.offsetLayer(layer, 2)) return;
		}
		
		//less than half layers
		for (int layer = threshold2; layer <= Util.view; layer++) {
			if (this.offsetLayer(layer, 1.5)) return;
		}
		
		//end with only full layers
		for (int layer = threshold1; layer <= Util.view; layer++) {
			if (this.offsetLayer(layer, 4)) return;
		}
		
		//not matched
		this.index = Integer.MAX_VALUE;
	}
	
	
}
