package com.bergerkiller.bukkit.nolagg;

import java.util.Set;
import java.util.logging.Level;

import net.minecraft.server.Block;
import net.minecraft.server.Chunk;
import net.minecraft.server.WorldServer;

@SuppressWarnings("rawtypes")
public class ChunkOperation implements Comparable {
	private int mode = 0;
	public final Chunk c;
	public Set ticks;
	public ChunkOperation(final Chunk chunk, Type type) {
		this.c = chunk;
		switch(type) {
		case LIGHTING : this.mode = 1; break;
		case UNLOAD : this.mode = 2; break;
		case AUTOSAVE : this.mode = 3; break;
		}
	}
	
	public static enum Type{LIGHTING, UNLOAD, AUTOSAVE};
	
	public int compareTo(Object object) {
		if (object instanceof ChunkOperation) {
			return ((ChunkOperation) object).mode - this.mode;
		} else {
			return 0;
		}
	}
	
	public int getMode() {
		return this.mode;
	}
		
	/*
	 * Whatever needs to happen when this object is being scheduled (synced)
	 */
	public void preexecute() {
		if (this.mode == 1) {
			//remove queued chunks for this chunk
		    ChunkSendQueue.deQueueGlobal(this.c);
		} else if (this.mode == 2 || this.mode == 3) {
			this.ticks = ChunkSaving.getTickList(this.getWorld());
		}
	}
		
	public WorldServer getWorld() {
		return (WorldServer) this.c.world;
	}
		
	public static boolean fixBlockLighting(final Chunk chunk, final int x, final int y, final int z) {
		final int typeid = chunk.getTypeId(x, y, z);
		if (!Block.o[typeid]) {
			final int factor = Math.max(1, Block.q[typeid]);
			final int light = chunk.h.a(x, y, z);
			//actual editing here
			int newlight = light + factor;
			if (x > 0) newlight = Math.max(newlight, chunk.h.a(x - 1, y, z));
			if (z > 0) newlight = Math.max(newlight, chunk.h.a(x, y, z - 1));
			if (y > 0) newlight = Math.max(newlight, chunk.h.a(x, y - 1, z));
			if (x < 15) newlight = Math.max(newlight, chunk.h.a(x + 1, y, z));
			if (z < 15) newlight = Math.max(newlight, chunk.h.a(x, y, z + 1));
			if (y < 127) newlight = Math.max(newlight, chunk.h.a(x, y + 1, z));
			newlight -= factor;
			//when exceeding in loops, just pick the highest value
			if (newlight > light) {
				chunk.h.a(x, y, z, newlight);
				return true;
			}
		}
		return false;
	}
	
	/*
	 * What needs to happen when the scheduler finds this object
	 */
	public void execute() {
		//1 = fix lighting
		//2 = save - unload mode
		//3 = save - autosave mode
		//save
		if (this.mode == 1) {
			c.h();
			c.initLighting();
			boolean haserror = true;
			int loops = -1;
			int x, y, z;
			while (haserror) {
				if (loops > 20) {
					NoLagg.log(Level.WARNING, "Failed to fix all lighting issues in chunk [" + c.x + "/" + c.z + "/" + c.world.getWorld().getName() + "]");
					break;
				}
				haserror = false;
				loops++;
				for (x = 0; x < 16; x++) {
					for (z = 0; z < 16; z++) {
						for (y = c.b(x,  z); y > 0; --y) {
							if (fixBlockLighting(c, x, y, z)) {
								haserror = true;
							}
						}
					}
				}
			}
			//queue
			ChunkSendQueue.scheduleQueueGlobal(this.c);
		} else if (this.mode == 2) {
			//unload
			ChunkSaving.saveChunk(this.c, this.ticks);
		} else if (this.mode == 3) {
			//autosave - detach entities!
			ChunkSaving.saveChunk(ChunkHandler.cloneChunk(this.c), this.ticks);
		}
	}
	
}
